@@ -19,7 +19,10 @@ class Options(optmanager.OptManager):
             "showhost",
             bool,
             False,
-            "Use the Host header to construct URLs for display.",
+            """Use the Host header to construct URLs for display.
+
+            This option is disabled by default because malicious apps may send misleading host headers to evade
+            your analysis. If this is not a concern, enable this options for better flow display.""",
         )
         self.add_option(
             "show_ignored_hosts",
@@ -72,18 +75,6 @@ class Options(optmanager.OptManager):
             process list. Specify it in config.yaml to avoid this.
             """,
         )
-        self.add_option(
-            "ciphers_client",
-            Optional[str],
-            None,
-            "Set supported ciphers for client <-> mitmproxy connections using OpenSSL syntax.",
-        )
-        self.add_option(
-            "ciphers_server",
-            Optional[str],
-            None,
-            "Set supported ciphers for mitmproxy <-> server connections using OpenSSL syntax.",
-        )
         self.add_option(
             "client_certs", Optional[str], None, "Client certificate file or directory."
         )
@@ -183,7 +174,10 @@ class Options(optmanager.OptManager):
             "ssl_insecure",
             bool,
             False,
-            "Do not verify upstream server SSL/TLS certificates.",
+            """Do not verify upstream server SSL/TLS certificates.
+
+            If this option is enabled, certificate validation is skipped and mitmproxy itself will be vulnerable to
+            TLS interception.""",
         )
         self.add_option(
             "ssl_verify_upstream_trusted_confdir",
